/*
  ==============================================================================

   This file is part of the JUCE library.
   Copyright (c) 2020 - Raw Material Software Limited

   JUCE is an open source library subject to commercial or open-source
   licensing.

   By using JUCE, you agree to the terms of both the JUCE 6 End-User License
   Agreement and JUCE Privacy Policy (both effective as of the 16th June 2020).

   End User License Agreement: www.juce.com/juce-6-licence
   Privacy Policy: www.juce.com/juce-privacy-policy

   Or: You may also use this code under the terms of the GPL v3 (see
   www.gnu.org/licenses).

   JUCE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY, AND ALL WARRANTIES, WHETHER
   EXPRESSED OR IMPLIED, INCLUDING MERCHANTABILITY AND FITNESS FOR PURPOSE, ARE
   DISCLAIMED.

  ==============================================================================
*/

#include "../JuceHeader.h"

namespace juce {

//==============================================================================
/**
    Reads and writes the lossless-compression FLAC audio format.

    To compile this, you'll need to set the JUCE_USE_FLAC flag.

    @see AudioFormat

    @tags{Audio}
*/
class JUCE_API PatchedFlacAudioFormat : public AudioFormat {
public:
  //==============================================================================
  PatchedFlacAudioFormat();
  ~PatchedFlacAudioFormat() override;

  //==============================================================================
  Array<int> getPossibleSampleRates() override;
  Array<int> getPossibleBitDepths() override;
  bool canDoStereo() override;
  bool canDoMono() override;
  bool isCompressed() override;
  StringArray getQualityOptions() override;

  //==============================================================================
  AudioFormatReader *createReaderFor(InputStream *sourceStream,
                                     bool deleteStreamIfOpeningFails) override;

  AudioFormatWriter *createWriterFor(OutputStream *streamToWriteTo,
                                     double sampleRateToUse,
                                     unsigned int numberOfChannels,
                                     int bitsPerSample,
                                     const StringPairArray &metadataValues,
                                     int qualityOptionIndex) override;
  using AudioFormat::createWriterFor;

private:
  JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR(PatchedFlacAudioFormat)
};

} // namespace juce
